@@ -11,7 +11,7 @@ fi
 set -e
 
 xandikos_cleanup() {
-	[ -z ${XANDIKOS_PID} ] || kill -TERM ${XANDIKOS_PID}
+	[ -z ${XANDIKOS_PID} ] || kill -INT ${XANDIKOS_PID}
 	rm --preserve-root -rf ${SERVEDIR}
 	cat ${DAEMON_LOG}
 	wait ${XANDIKOS_PID} || true
