@@ -22,9 +22,13 @@
 https://tools.ietf.org/html/rfc4791
 """
 import datetime
-import logging
+import itertools
 import pytz
 
+from .icalendar import (
+    apply_time_range_vevent,
+    as_tz_aware_ts,
+)
 from icalendar.cal import component_factory, Calendar as ICalendar, FreeBusy
 from icalendar.prop import vDDDTypes, vPeriod, LocalTimezone
 
@@ -41,6 +45,8 @@ NAMESPACE = 'urn:ietf:params:xml:ns:caldav'
 # https://tools.ietf.org/html/rfc4791, section 4.2
 CALENDAR_RESOURCE_TYPE = '{%s}calendar' % NAMESPACE
 
+SUBSCRIPTION_RESOURCE_TYPE = '{http://calendarserver.org/ns/}subscribed'
+
 # TODO(jelmer): These resource types belong in scheduling.py
 SCHEDULE_INBOX_RESOURCE_TYPE = '{%s}schedule-inbox' % NAMESPACE
 SCHEDULE_OUTBOX_RESOURCE_TYPE = '{%s}schedule-outbox' % NAMESPACE
@@ -52,32 +58,32 @@ TRANSPARENCY_TRANSPARENT = 'transparent'
 TRANSPARENCY_OPAQUE = 'opaque'
 
 
-class MissingProperty(Exception):
-
-    def __init__(self, property_name):
-        super(MissingProperty, self).__init__(
-            "Property %r missing" % property_name)
-        self.property_name = property_name
-
-
 class Calendar(webdav.Collection):
 
     resource_types = (webdav.Collection.resource_types +
                       [CALENDAR_RESOURCE_TYPE])
 
-    def get_calendar_description(self):
+    def get_calendar_description(self) -> str:
         """Return the calendar description."""
         raise NotImplementedError(self.get_calendar_description)
 
-    def get_calendar_color(self):
+    def get_calendar_color(self) -> str:
         """Return the calendar color."""
         raise NotImplementedError(self.get_calendar_color)
 
-    def set_calendar_color(self, color):
+    def set_calendar_color(self, color: str) -> None:
         """Set the calendar color."""
         raise NotImplementedError(self.set_calendar_color)
 
-    def get_calendar_timezone(self):
+    def get_calendar_order(self) -> str:
+        """Return the calendar order."""
+        raise NotImplementedError(self.get_calendar_order)
+
+    def set_calendar_order(self, order: str) -> None:
+        """Set the calendar order."""
+        raise NotImplementedError(self.set_calendar_order)
+
+    def get_calendar_timezone(self) -> str:
         """Return calendar timezone.
 
         This should be an iCalendar object with exactly one
@@ -85,7 +91,7 @@ class Calendar(webdav.Collection):
         """
         raise NotImplementedError(self.get_calendar_timezone)
 
-    def set_calendar_timezone(self):
+    def set_calendar_timezone(self, content: str) -> None:
         """Set calendar timezone.
 
         This should be an iCalendar object with exactly one
@@ -93,14 +99,14 @@ class Calendar(webdav.Collection):
         """
         raise NotImplementedError(self.set_calendar_timezone)
 
-    def get_supported_calendar_components(self):
+    def get_supported_calendar_components(self) -> str:
         """Return set of supported calendar components in this calendar.
 
         :return: iterable over component names
         """
         raise NotImplementedError(self.get_supported_calendar_components)
 
-    def get_supported_calendar_data_types(self):
+    def get_supported_calendar_data_types(self) -> str:
         """Return supported calendar data types.
 
         :return: iterable over (content_type, version) tuples
@@ -139,10 +145,6 @@ class Calendar(webdav.Collection):
         """Return max attachment size."""
         raise NotImplementedError(self.get_max_attachment_size)
 
-    def get_managed_attachments_server_url(self):
-        """Return the attachments server URL."""
-        raise NotImplementedError(self.get_managed_attachments_server_url)
-
     def get_schedule_calendar_transparency(self):
         """Get calendar transparency.
 
@@ -150,6 +152,58 @@ class Calendar(webdav.Collection):
         """
         return TRANSPARENCY_OPAQUE
 
+    def calendar_query(self, create_filter_fn):
+        """Query for all the members of this calendar that match `filter`.
+
+        This is a naive implementation; subclasses should ideally provide
+        their own implementation that is faster.
+
+        :param create_filter_fn: Callback that constructs a
+            filter; takes a filter building class.
+        :return: Iterator over name, resource objects
+        """
+        raise NotImplementedError(self.calendar_query)
+
+
+class Subscription(object):
+
+    resource_types = (webdav.Collection.resource_types +
+                      [SUBSCRIPTION_RESOURCE_TYPE])
+
+    def get_source_url(self):
+        """Get the source URL for this calendar."""
+        raise NotImplementedError(self.get_source_url)
+
+    def set_source_url(self, url):
+        """Set the source URL for this calendar."""
+        raise NotImplementedError(self.set_source_url)
+
+    def get_calendar_description(self):
+        """Return the calendar description."""
+        raise NotImplementedError(self.get_calendar_description)
+
+    def get_calendar_color(self):
+        """Return the calendar color."""
+        raise NotImplementedError(self.get_calendar_color)
+
+    def set_calendar_color(self, color):
+        """Set the calendar color."""
+        raise NotImplementedError(self.set_calendar_color)
+
+    def get_supported_calendar_components(self):
+        """Return set of supported calendar components in this calendar.
+
+        :return: iterable over component names
+        """
+        raise NotImplementedError(self.get_supported_calendar_components)
+
+
+class CalendarHomeSet(object):
+
+    def get_managed_attachments_server_url(self):
+        """Return the attachments server URL."""
+        raise NotImplementedError(self.get_managed_attachments_server_url)
+
 
 class PrincipalExtensions:
     """CalDAV-specific extensions to DAVPrincipal."""
@@ -180,7 +234,7 @@ class CalendarHomeSetProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_home_set():
             href = webdav.ensure_trailing_slash(href)
             el.append(webdav.create_href(href, base_href))
@@ -193,16 +247,38 @@ class CalendarDescriptionProperty(webdav.Property):
     """
 
     name = '{%s}calendar-description' % NAMESPACE
-    resource_type = CALENDAR_RESOURCE_TYPE
+    resource_type = (CALENDAR_RESOURCE_TYPE, SUBSCRIPTION_RESOURCE_TYPE)
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         el.text = resource.get_calendar_description()
 
     # TODO(jelmer): allow modification of this property
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         raise NotImplementedError
 
 
+def _extract_from_component(incomp, outcomp, requested):
+    for tag in requested:
+        if tag.tag == ('{%s}comp' % NAMESPACE):
+            for insub in incomp.subcomponents:
+                if insub.name == tag.get('name'):
+                    outsub = component_factory[insub.name]()
+                    outcomp.add_component(outsub)
+                    _extract_from_component(insub, outsub, tag)
+        elif tag.tag == ('{%s}prop' % NAMESPACE):
+            outcomp[tag.get('name')] = incomp[tag.get('name')]
+        elif tag.tag == ('{%s}allprop' % NAMESPACE):
+            for propname in incomp:
+                outcomp[propname] = incomp[propname]
+        elif tag.tag == ('{%s}allcomp' % NAMESPACE):
+            for insub in incomp.subcomponents:
+                outsub = component_factory[insub.name]()
+                outcomp.add_component(outsub)
+                _extract_from_component(insub, outsub, tag)
+        else:
+            raise AssertionError('invalid element %r' % tag)
+
+
 def extract_from_calendar(incal, outcal, requested):
     """Extract requested components/properties from calendar.
 
@@ -210,17 +286,22 @@ def extract_from_calendar(incal, outcal, requested):
     :param outcal: Calendar to write to
     :param requested: <calendar-data> element with requested
         components/properties
-    :return: A Calendar
     """
     for tag in requested:
         if tag.tag == ('{%s}comp' % NAMESPACE):
-            for insub in incal.subcomponents:
-                if insub.name == tag.get('name'):
-                    outsub = component_factory[insub.name]
-                    outcal.add_component(outsub)
-                    extract_from_calendar(insub, outsub, tag)
-        elif tag.tag == ('{%s}prop' % NAMESPACE):
-            outcal[tag.get('name')] = incal[tag.get('name')]
+            if incal.name == tag.get('name'):
+                _extract_from_component(incal, outcal, tag)
+        elif tag.tag == ('{%s}expand' % NAMESPACE):
+            # TODO(jelmer): https://github.com/jelmer/xandikos/issues/102
+            raise NotImplementedError('expand is not yet implemented')
+        elif tag.tag == ('{%s}limit-recurrence-set' % NAMESPACE):
+            # TODO(jelmer): https://github.com/jelmer/xandikos/issues/103
+            raise NotImplementedError(
+                'limit-recurrence-set is not yet implemented')
+        elif tag.tag == ('{%s}limit-freebusy-set' % NAMESPACE):
+            # TODO(jelmer): https://github.com/jelmer/xandikos/issues/104
+            raise NotImplementedError(
+                'limit-freebusy-set is not yet implemented')
         else:
             raise AssertionError('invalid element %r' % tag)
 
@@ -239,9 +320,9 @@ class CalendarDataProperty(davcommon.SubbedProperty):
     def supported_on(self, resource):
         return (resource.get_content_type() == 'text/calendar')
 
-    def get_value_ext(self, base_href, resource, el, environ, requested):
+    async def get_value_ext(self, base_href, resource, el, environ, requested):
         if len(requested) == 0:
-            serialized_cal = b''.join(resource.get_body())
+            serialized_cal = b''.join(await resource.get_body())
         else:
             c = ICalendar()
             calendar = calendar_from_resource(resource)
@@ -254,6 +335,20 @@ class CalendarDataProperty(davcommon.SubbedProperty):
         el.text = serialized_cal.decode('utf-8')
 
 
+class CalendarOrderProperty(webdav.Property):
+    """Provides calendar-order property.
+    """
+
+    name = '{http://apple.com/ns/ical/}calendar-order'
+    resource_type = CALENDAR_RESOURCE_TYPE
+
+    async def get_value(self, base_href, resource, el, environ):
+        el.text = resource.get_calendar_order()
+
+    async def set_value(self, href, resource, el):
+        resource.set_calendar_order(el.text)
+
+
 class CalendarMultiGetReporter(davcommon.MultiGetReporter):
 
     name = '{%s}calendar-multiget' % NAMESPACE
@@ -261,69 +356,52 @@ class CalendarMultiGetReporter(davcommon.MultiGetReporter):
     data_property = CalendarDataProperty()
 
 
-def apply_prop_filter(el, comp, tzify):
+def parse_prop_filter(el, cls):
     name = el.get('name')
+
     # From https://tools.ietf.org/html/rfc4791, 9.7.2:
     # A CALDAV:comp-filter is said to match if:
 
-    # The CALDAV:prop-filter XML element contains a CALDAV:is-not-defined XML
-    # element and no property of the type specified by the "name" attribute
-    # exists in the enclosing calendar component;
-    if (
-        len(el) == 1 and
-        el[0].tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined'
-    ):
-        return name not in comp
-
-    try:
-        prop = comp[name]
-    except KeyError:
-        return False
+    prop_filter = cls(name=name)
 
     for subel in el:
-        if subel.tag == '{urn:ietf:params:xml:ns:caldav}time-range':
-            if not apply_time_range_prop(subel, prop, tzify):
-                return False
+        if subel.tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined':
+            prop_filter.is_not_defined = True
+        elif subel.tag == '{urn:ietf:params:xml:ns:caldav}time-range':
+            parse_time_range(subel, prop_filter.filter_time_range)
         elif subel.tag == '{urn:ietf:params:xml:ns:caldav}text-match':
-            if not apply_text_match(subel, prop):
-                return False
+            parse_text_match(subel, prop_filter.filter_text_match)
         elif subel.tag == '{urn:ietf:params:xml:ns:caldav}param-filter':
-            if not apply_param_filter(subel, prop):
-                return False
-    return True
+            parse_param_filter(subel, prop_filter.filter_parameter)
+        elif subel.tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined':
+            pass
+        else:
+            raise AssertionError("unknown subelement %r" % subel.tag)
+    return prop_filter
 
 
-def apply_text_match(el, value):
+def parse_text_match(el, cls):
     collation = el.get('collation', 'i;ascii-casemap')
     negate_condition = el.get('negate-condition', 'no')
-    matches = davcommon.get_collation(collation)(el.text, value)
 
-    if negate_condition == 'yes':
-        return (not matches)
-    else:
-        return matches
+    return cls(
+        el.text, collation=collation,
+        negate_condition=(negate_condition == 'yes'))
 
 
-def apply_param_filter(el, prop):
+def parse_param_filter(el, cls):
     name = el.get('name')
-    if (
-        len(el) == 1 and
-        el[0].tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined'
-    ):
-        return name not in prop.params
 
-    try:
-        value = prop.params[name]
-    except KeyError:
-        return False
+    param_filter = cls(name=name)
 
     for subel in el:
-        if subel.tag == '{urn:ietf:params:xml:ns:caldav}text-match':
-            if not apply_text_match(subel, value):
-                return False
+        if subel.tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined':
+            param_filter.is_not_defined = True
+        elif subel.tag == '{urn:ietf:params:xml:ns:caldav}text-match':
+            parse_text_match(subel, param_filter.filter_time_range)
         else:
             raise AssertionError('unknown tag %r in param-filter', subel.tag)
-    return True
+    return param_filter
 
 
 def _parse_time_range(el):
@@ -344,150 +422,20 @@ def _parse_time_range(el):
     return (start, end)
 
 
-def as_tz_aware_ts(dt, default_timezone):
-    if not getattr(dt, 'time', None):
-        dt = datetime.datetime.combine(dt, datetime.time())
-    if dt.tzinfo is None:
-        dt = dt.replace(tzinfo=default_timezone)
-    assert dt.tzinfo
-    return dt
-
-
-def apply_time_range_vevent(start, end, comp, tzify):
-    if comp['DTSTART'] is None:
-        raise MissingProperty('DTSTART')
-
-    if not (end > tzify(comp['DTSTART'].dt)):
-        return False
-
-    if 'DTEND' in comp:
-        if tzify(comp['DTEND'].dt) < tzify(comp['DTSTART'].dt):
-            logging.debug('Invalid DTEND < DTSTART')
-        return (start < tzify(comp['DTEND'].dt))
-
-    if 'DURATION' in comp:
-        return (start < tzify(comp['DTSTART'].dt) + comp['DURATION'].dt)
-    if getattr(comp['DTSTART'].dt, 'time', None) is not None:
-        return (start <= tzify(comp['DTSTART'].dt))
-    else:
-        return (start < (tzify(comp['DTSTART'].dt) + datetime.timedelta(1)))
-
-
-def apply_time_range_vjournal(start, end, comp, tzify):
-    if 'DTSTART' not in comp:
-        return False
-
-    if not (end > tzify(comp['DTSTART'].dt)):
-        return False
-
-    if getattr(comp['DTSTART'].dt, 'time', None) is not None:
-        return (start <= tzify(comp['DTSTART'].dt))
-    else:
-        return (start < (tzify(comp['DTSTART'].dt) + datetime.timedelta(1)))
-
-
-def apply_time_range_vtodo(start, end, comp, tzify):
-    if 'DTSTART' in comp:
-        if 'DURATION' in comp and 'DUE' not in comp:
-            return (
-                start <= tzify(comp['DTSTART'].dt) + comp['DURATION'].dt and
-                (end > tzify(comp['DTSTART'].dt) or
-                 end >= tzify(comp['DTSTART'].dt) + comp['DURATION'].dt)
-            )
-        elif 'DUE' in comp and 'DURATION' not in comp:
-            return (
-                (start <= tzify(comp['DTSTART'].dt) or
-                 start < tzify(comp['DUE'].dt)) and
-                (end > tzify(comp['DTSTART'].dt) or
-                 end < tzify(comp['DUE'].dt))
-            )
-        else:
-            return (start <= tzify(comp['DTSTART'].dt) and
-                    end > tzify(comp['DTSTART'].dt))
-    elif 'DUE' in comp:
-        return start < tzify(comp['DUE'].dt) and end >= tzify(comp['DUE'].dt)
-    elif 'COMPLETED' in comp:
-        if 'CREATED' in comp:
-            return (
-                (start <= tzify(comp['CREATED'].dt) or
-                 start <= tzify(comp['COMPLETED'].dt)) and
-                (end >= tzify(comp['CREATED'].dt) or
-                 end >= tzify(comp['COMPLETED'].dt))
-            )
-        else:
-            return (
-                start <= tzify(comp['COMPLETED'].dt) and
-                end >= tzify(comp['COMPLETED'].dt)
-            )
-    elif 'CREATED' in comp:
-        return end >= tzify(comp['CREATED'].dt)
-    else:
-        return True
-
-
-def apply_time_range_vfreebusy(start, end, comp, tzify):
-    if 'DTSTART' in comp and 'DTEND' in comp:
-        return (
-            start <= tzify(comp['DTEND'].dt) and
-            end > tzify(comp['DTEND'].dt)
-        )
-
-    for period in comp.get('FREEBUSY', []):
-        if start < period.end and end > period.start:
-            return True
-
-    return False
-
-
-def apply_time_range_valarm(start, end, comp, tzify):
-    raise NotImplementedError(apply_time_range_valarm)
-
-
-def apply_time_range_comp(el, comp, tzify):
-    # According to https://tools.ietf.org/html/rfc4791, section 9.9 these are
-    # the properties to check.
+def parse_time_range(el, cls):
     (start, end) = _parse_time_range(el)
-    component_handlers = {
-        'VEVENT': apply_time_range_vevent,
-        'VTODO': apply_time_range_vtodo,
-        'VJOURNAL': apply_time_range_vjournal,
-        'VFREEBUSY': apply_time_range_vfreebusy,
-        'VALARM': apply_time_range_valarm}
-    try:
-        component_handler = component_handlers[comp.name]
-    except KeyError:
-        logging.warning('unknown component %r in time-range filter',
-                        comp.name)
-        return False
-    return component_handler(start, end, comp, tzify)
+    return cls(start, end)
 
 
-def apply_time_range_prop(el, val, tzify):
-    (start, end) = _parse_time_range(el)
-    raise NotImplementedError(apply_time_range_prop)
-
-
-def apply_comp_filter(el, comp, tzify):
+def parse_comp_filter(el: ET.Element, cls):
     """Compile a comp-filter element into a Python function.
     """
     name = el.get('name')
+
     # From https://tools.ietf.org/html/rfc4791, 9.7.1:
     # A CALDAV:comp-filter is said to match if:
 
-    # 2. The CALDAV:comp-filter XML element contains a CALDAV:is-not-defined
-    # XML element and the calendar object or calendar component type specified
-    # by the "name" attribute does not exist in the current scope;
-    if (
-        len(el) == 1 and
-        el[0].tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined'
-    ):
-        return comp.name != name
-
-    # 1: The CALDAV:comp-filter XML element is empty and the calendar object or
-    # calendar component type specified by the "name" attribute exists in the
-    # current scope;
-    if comp.name != name:
-        return False
+    comp_filter = cls(name=name)
 
     # 3. The CALDAV:comp-filter XML element contains a CALDAV:time-range XML
     # element and at least one recurrence instance in the targeted calendar
@@ -495,19 +443,26 @@ def apply_comp_filter(el, comp, tzify):
     # specified CALDAV:prop-filter and CALDAV:comp-filter child XML elements
     # also match the targeted calendar component;
     for subel in el:
+        if subel.tag == '{urn:ietf:params:xml:ns:caldav}is-not-defined':
+            comp_filter.is_not_defined = True
         if subel.tag == '{urn:ietf:params:xml:ns:caldav}comp-filter':
-            if not any(apply_comp_filter(subel, c, tzify)
-                       for c in comp.subcomponents):
-                return False
+            parse_comp_filter(subel, comp_filter.filter_subcomponent)
         elif subel.tag == '{urn:ietf:params:xml:ns:caldav}prop-filter':
-            if not apply_prop_filter(subel, comp, tzify):
-                return False
+            parse_prop_filter(subel, comp_filter.filter_property)
         elif subel.tag == '{urn:ietf:params:xml:ns:caldav}time-range':
-            if not apply_time_range_comp(subel, comp, tzify):
-                return False
+            parse_time_range(subel, comp_filter.filter_time_range)
         else:
             raise AssertionError('unknown filter tag %r' % subel.tag)
-    return True
+    return comp_filter
+
+
+def parse_filter(filter_el: ET.Element, cls):
+    for subel in filter_el:
+        if subel.tag == '{urn:ietf:params:xml:ns:caldav}comp-filter':
+            parse_comp_filter(subel, cls.filter_subcomponent)
+        else:
+            raise AssertionError('unknown filter tag %r' % subel.tag)
+    return cls
 
 
 def calendar_from_resource(resource):
@@ -519,18 +474,6 @@ def calendar_from_resource(resource):
     return resource.file.calendar
 
 
-def apply_filter(el, resource, tzify):
-    """Compile a filter element into a Python function.
-    """
-    if el is None:
-        # Empty filter, let's not bother parsing
-        return lambda x: True
-    c = calendar_from_resource(resource)
-    if c is None:
-        return False
-    return apply_comp_filter(list(el)[0], c, tzify)
-
-
 def extract_tzid(cal):
     return cal.subcomponents[0]['TZID']
 
@@ -556,9 +499,10 @@ class CalendarQueryReporter(webdav.Reporter):
     data_property = CalendarDataProperty()
 
     @webdav.multistatus
-    def report(self, environ, body, resources_by_hrefs, properties, base_href,
-               base_resource, depth):
-        # TODO(jelmer): Verify that resource is an addressbook
+    async def report(
+            self, environ, body, resources_by_hrefs, properties, base_href,
+            base_resource, depth):
+        # TODO(jelmer): Verify that resource is a calendar
         requested = None
         filter_el = None
         tztext = None
@@ -577,23 +521,24 @@ class CalendarQueryReporter(webdav.Reporter):
         else:
             tz = get_calendar_timezone(base_resource)
 
-        def tzify(dt):
-            return as_tz_aware_ts(dt, tz)
-        for (href, resource) in webdav.traverse_resource(
-                base_resource, base_href, depth):
-            try:
-                filter_result = apply_filter(filter_el, resource, tzify)
-            except MissingProperty as e:
-                logging.warning(
-                    'calendar_query: Ignoring calendar object %s, due '
-                    'to missing property %s', href, e.property_name)
-                continue
-            if not filter_result:
-                continue
-            propstat = davcommon.get_properties_with_data(
-                self.data_property, href, resource, properties, environ,
-                requested)
-            yield webdav.Status(href, '200 OK', propstat=list(propstat))
+        def filter_fn(cls):
+            return parse_filter(filter_el, cls(tz))
+
+        def members(collection):
+            return itertools.chain(
+                collection.calendar_query(filter_fn),
+                collection.subcollections())
+
+        async for (href, resource) in webdav.traverse_resource(
+                base_resource, base_href, depth,
+                members=members):
+            # Ideally traverse_resource would only return the right things.
+            if getattr(resource, 'content_type', None) == 'text/calendar':
+                propstat = davcommon.get_properties_with_data(
+                    self.data_property, href, resource, properties, environ,
+                    requested)
+                yield webdav.Status(
+                    href, '200 OK', propstat=[s async for s in propstat])
 
 
 class CalendarColorProperty(webdav.Property):
@@ -603,12 +548,12 @@ class CalendarColorProperty(webdav.Property):
     """
 
     name = '{http://apple.com/ns/ical/}calendar-color'
-    resource_type = CALENDAR_RESOURCE_TYPE
+    resource_type = (CALENDAR_RESOURCE_TYPE, SUBSCRIPTION_RESOURCE_TYPE)
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_calendar_color()
 
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         resource.set_calendar_color(el.text)
 
 
@@ -623,11 +568,12 @@ class SupportedCalendarComponentSetProperty(webdav.Property):
     name = '{%s}supported-calendar-component-set' % NAMESPACE
     resource_type = (CALENDAR_RESOURCE_TYPE,
                      SCHEDULE_INBOX_RESOURCE_TYPE,
-                     SCHEDULE_OUTBOX_RESOURCE_TYPE)
+                     SCHEDULE_OUTBOX_RESOURCE_TYPE,
+                     SUBSCRIPTION_RESOURCE_TYPE)
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         for component in resource.get_supported_calendar_components():
             subel = ET.SubElement(el, '{urn:ietf:params:xml:ns:caldav}comp')
             subel.set('name', component)
@@ -645,7 +591,7 @@ class SupportedCalendarDataProperty(webdav.Property):
                      SCHEDULE_OUTBOX_RESOURCE_TYPE)
     in_allprops = False
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         for (content_type, version) in (
                 resource.get_supported_calendar_data_types()):
             subel = ET.SubElement(
@@ -665,10 +611,10 @@ class CalendarTimezoneProperty(webdav.Property):
                      SCHEDULE_INBOX_RESOURCE_TYPE)
     in_allprops = False
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_calendar_timezone()
 
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         if el is not None:
             resource.set_calendar_timezone(el.text)
         else:
@@ -688,7 +634,7 @@ class MinDateTimeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_min_date_time()
 
 
@@ -705,7 +651,7 @@ class MaxDateTimeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_max_date_time()
 
 
@@ -721,7 +667,7 @@ class MaxInstancesProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_instances())
 
 
@@ -738,7 +684,7 @@ class MaxAttendeesPerInstanceProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attendees_per_instance())
 
 
@@ -755,7 +701,7 @@ class MaxResourceSizeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_resource_size())
 
 
@@ -770,7 +716,7 @@ class MaxAttachmentsPerResourceProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attachments_per_resource())
 
 
@@ -785,7 +731,7 @@ class MaxAttachmentSizeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attachment_size())
 
 
@@ -798,12 +744,35 @@ class ManagedAttachmentsServerURLProperty(webdav.Property):
     name = '{%s}managed-attachments-server-URL' % NAMESPACE
     in_allprops = False
 
-    def get_value(self, base_href, resource, el, environ):
-        href = resource.get_managed_attachments_server_url()
+    async def get_value(self, base_href, resource, el, environ):
+        # The RFC specifies that this property can be set on a calendar home
+        # collection.
+        # However, there is no matching resource type and we don't want to
+        # force all resources to implement it. So we just check whether the
+        # attribute is present.
+        fn = getattr(resource, 'get_managed_attachments_server_url', None)
+        if fn is None:
+            raise KeyError
+        href = fn()
         if href is not None:
             el.append(webdav.create_href(href, base_href))
 
 
+class SourceProperty(webdav.Property):
+    """source property.
+    """
+    name = '{http://calendarserver.org/ns/}source'
+    resource_type = SUBSCRIPTION_RESOURCE_TYPE
+    in_allprops = True
+    live = False
+
+    async def get_value(self, base_href, resource, el, environ):
+        el.append(webdav.create_href(resource.get_source_url(), base_href))
+
+    async def set_value(self, href, resource, el):
+        raise NotImplementedError(self.set_value)
+
+
 class CalendarProxyReadForProperty(webdav.Property):
     """calendar-proxy-read-for property.
 
@@ -816,7 +785,7 @@ class CalendarProxyReadForProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_proxy_read_for():
             el.append(webdav.create_href(href, base_href))
 
@@ -833,7 +802,7 @@ class CalendarProxyWriteForProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_proxy_write_for():
             el.append(webdav.create_href(href, base_href))
 
@@ -848,7 +817,7 @@ class ScheduleCalendarTransparencyProperty(webdav.Property):
     live = False
     resource_type = CALENDAR_RESOURCE_TYPE
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         transp = resource.get_schedule_calendar_transparency()
         if transp == TRANSPARENCY_TRANSPARENT:
             ET.SubElement(el, '{%s}transparent' % NAMESPACE)
@@ -889,8 +858,8 @@ def extract_freebusy(comp, tzify):
     return ret
 
 
-def iter_freebusy(resources, start, end, tzify):
-    for (href, resource) in resources:
+async def iter_freebusy(resources, start, end, tzify):
+    async for (href, resource) in resources:
         c = calendar_from_resource(resource)
         if c is None:
             continue
@@ -913,8 +882,8 @@ class FreeBusyQueryReporter(webdav.Reporter):
     name = '{urn:ietf:params:xml:ns:caldav}free-busy-query'
     resource_type = CALENDAR_RESOURCE_TYPE
 
-    def report(self, environ, start_response, body, resources_by_hrefs,
-               properties, base_href, base_resource, depth):
+    async def report(self, environ, body, resources_by_hrefs,
+                     properties, base_href, base_resource, depth):
         requested = None
         for el in body:
             if el.tag == '{urn:ietf:params:xml:ns:caldav}time-range':
@@ -935,47 +904,42 @@ class FreeBusyQueryReporter(webdav.Reporter):
         fb['DTSTAMP'] = vDDDTypes(tzify(datetime.datetime.now()))
         fb['DTSTART'] = vDDDTypes(start)
         fb['DTEND'] = vDDDTypes(end)
-        fb['FREEBUSY'] = list(iter_freebusy(
+        fb['FREEBUSY'] = [item async for item in iter_freebusy(
             webdav.traverse_resource(base_resource, base_href, depth),
-            start, end, tzify))
+            start, end, tzify)]
         ret.add_component(fb)
-        start_response('200 OK', [])
-        return [ret.to_ical()]
+        return webdav.Response(status='200 OK', body=[ret.to_ical()])
 
 
 class MkcalendarMethod(webdav.Method):
 
-    def handle(self, environ, start_response, app):
-        try:
-            content_type = environ['CONTENT_TYPE']
-        except KeyError:
-            base_content_type = None
-        else:
-            base_content_type, params = webdav.parse_type(content_type)
+    async def handle(self, request, environ, app):
+        content_type = request.content_type
+        base_content_type, params = webdav.parse_type(content_type)
         if base_content_type not in (
-            'text/xml', 'application/xml', None, 'text/plain'
+            'text/xml', 'application/xml', None, 'text/plain',
+            'application/octet-stream',
         ):
             raise webdav.UnsupportedMediaType(content_type)
-        href, path, resource = app._get_resource_from_environ(environ)
+        href, path, resource = app._get_resource_from_environ(request, environ)
         if resource is not None:
             return webdav._send_simple_dav_error(
-                environ, start_response,
+                request,
                 '403 Forbidden',
                 error=ET.Element('{DAV:}resource-must-be-null'),
                 description=('Something already exists at %r' % path))
         try:
             resource = app.backend.create_collection(path)
         except FileNotFoundError:
-            start_response('409 Conflict', [])
-            return []
+            return webdav.Response(status='409 Conflict')
         el = ET.Element('{DAV:}resourcetype')
-        app.properties['{DAV:}resourcetype'].get_value(
+        await app.properties['{DAV:}resourcetype'].get_value(
             href, resource, el, environ)
         ET.SubElement(el, '{urn:ietf:params:xml:ns:caldav}calendar')
         app.properties['{DAV:}resourcetype'].set_value(href, resource, el)
         if base_content_type in ('text/xml', 'application/xml'):
-            et = webdav._readXmlBody(
-                environ, '{urn:ietf:params:xml:ns:caldav}mkcalendar')
+            et = await webdav._readXmlBody(
+                request, '{urn:ietf:params:xml:ns:caldav}mkcalendar')
             propstat = []
             for el in et:
                 if el.tag != '{DAV:}set':
@@ -988,7 +952,6 @@ class MkcalendarMethod(webdav.Method):
             for propstat_el in webdav.propstat_as_xml(propstat):
                 ret.append(propstat_el)
             return webdav._send_xml_response(
-                start_response, '201 Created', ret, webdav.DEFAULT_ENCODING)
+                '201 Created', ret, webdav.DEFAULT_ENCODING)
         else:
-            start_response('201 Created', [])
-            return []
+            return webdav.Response(status='201 Created')
