@@ -24,11 +24,12 @@ def test_suite():
     names = [
         'api',
         'caldav',
-        'collectionconfig',
+        'config',
         'icalendar',
         'store',
         'webdav',
         'web',
+        'wsgi',
     ]
     module_names = ['xandikos.tests.test_' + name for name in names]
     loader = unittest.TestLoader()
