@@ -27,9 +27,11 @@ import unittest
 from dulwich.objects import Blob, Commit, Tree
 from dulwich.repo import Repo
 
+from typing import Type
+
 from xandikos.icalendar import ICalendarFile
 from xandikos.store import (
-    DuplicateUidError, File, InvalidETag, NoSuchItem)
+    DuplicateUidError, File, InvalidETag, NoSuchItem, Filter, Store)
 from xandikos.store.git import (
     GitStore, BareGitStore, TreeGitStore)
 from xandikos.store.vdir import (
@@ -120,6 +122,91 @@ class BaseStoreTest(object):
         self.assertEqual([('foo.ics', 'text/calendar', etag)],
                          list(gc.iter_with_etag()))
 
+    def test_with_filter(self):
+        gc = self.create_store()
+        (name1, etag1) = gc.import_one('foo.ics', 'text/calendar',
+                                       [EXAMPLE_VCALENDAR1])
+        (name2, etag2) = gc.import_one('bar.ics', 'text/calendar',
+                                       [EXAMPLE_VCALENDAR2])
+
+        class DummyFilter(Filter):
+
+            content_type = 'text/calendar'
+
+            def __init__(self, text):
+                self.text = text
+
+            def check(self, name, resource):
+                return self.text in b''.join(resource.content)
+
+        self.assertEqual(
+            2, len(list(gc.iter_with_filter(
+                filter=DummyFilter(b'do something')))))
+
+        [(ret_name, ret_file, ret_etag)] = list(gc.iter_with_filter(
+            filter=DummyFilter(b'do something else')))
+        self.assertEqual(ret_name, name2)
+        self.assertEqual(ret_etag, etag2)
+        self.assertEqual(ret_file.content_type, 'text/calendar')
+        self.assertEqual(
+            b''.join(ret_file.content),
+            EXAMPLE_VCALENDAR2.replace(b'\n', b'\r\n'))
+
+    def test_get_by_index(self):
+        gc = self.create_store()
+        (name1, etag1) = gc.import_one('foo.ics', 'text/calendar',
+                                       [EXAMPLE_VCALENDAR1])
+        (name2, etag2) = gc.import_one('bar.ics', 'text/calendar',
+                                       [EXAMPLE_VCALENDAR2])
+        (name3, etag3) = gc.import_one('bar.txt', 'text/plain',
+                                       [b'Not a calendar file.'])
+        self.assertEqual({}, dict(gc.index_manager.desired))
+
+        filtertext = 'C=VCALENDAR/C=VTODO/P=SUMMARY'
+
+        class DummyFilter(Filter):
+
+            content_type = 'text/calendar'
+
+            def __init__(self, text):
+                self.text = text
+
+            def index_keys(self):
+                return [[filtertext]]
+
+            def check_from_indexes(self, name, index_values):
+                return any(self.text in v.encode()
+                           for v in index_values[filtertext])
+
+            def check(self, name, resource):
+                return self.text in b''.join(resource.content)
+
+        self.assertEqual(
+            2, len(list(gc.iter_with_filter(
+                filter=DummyFilter(b'do something')))))
+
+        [(ret_name, ret_file, ret_etag)] = list(gc.iter_with_filter(
+            filter=DummyFilter(b'do something else')))
+        self.assertEqual(
+            {filtertext: 2},
+            dict(gc.index_manager.desired))
+
+        # Force index
+        gc.index.reset([filtertext])
+
+        [(ret_name, ret_file, ret_etag)] = list(gc.iter_with_filter(
+            filter=DummyFilter(b'do something else')))
+        self.assertEqual(
+            {filtertext: 2},
+            dict(gc.index_manager.desired))
+
+        self.assertEqual(ret_name, name2)
+        self.assertEqual(ret_etag, etag2)
+        self.assertEqual(ret_file.content_type, 'text/calendar')
+        self.assertEqual(
+            b''.join(ret_file.content),
+            EXAMPLE_VCALENDAR2.replace(b'\n', b'\r\n'))
+
     def test_import_one_duplicate_uid(self):
         gc = self.create_store()
         (name, etag) = gc.import_one('foo.ics', 'text/calendar',
@@ -230,7 +317,7 @@ class VdirStoreTest(BaseStoreTest, unittest.TestCase):
 
 class BaseGitStoreTest(BaseStoreTest):
 
-    kls = None
+    kls: Type[Store]
 
     def create_store(self):
         raise NotImplementedError(self.create_store)
@@ -262,13 +349,13 @@ class BaseGitStoreTest(BaseStoreTest):
             [('foo.ics', 'text/calendar', bid)],
             list(gc.iter_with_etag()))
 
-    def test_get_description(self):
+    def test_get_description_from_git_config(self):
         gc = self.create_store()
-        try:
-            gc.repo.set_description(b'a repo description')
-        except NotImplementedError:
-            self.skipTest('old dulwich version without '
-                          'MemoryRepo.set_description')
+        config = gc.repo.get_config()
+        config.set(b'xandikos', b'test', b'test')
+        if getattr(config, 'path', None):
+            config.write_to_path()
+        gc.repo.set_description(b'a repo description')
         self.assertEqual(gc.get_description(), 'a repo description')
 
     def test_displayname(self):
@@ -289,6 +376,15 @@ class BaseGitStoreTest(BaseStoreTest):
             c.write_to_path()
         self.assertEqual('334433', gc.get_color())
 
+    def test_get_source_url(self):
+        gc = self.create_store()
+        self.assertIs(None, gc.get_source_url())
+        c = gc.repo.get_config()
+        c.set(b'xandikos', b'source', b'www.google.com')
+        if getattr(c, 'path', None):
+            c.write_to_path()
+        self.assertEqual('www.google.com', gc.get_source_url())
+
     def test_default_no_subdirectories(self):
         gc = self.create_store()
         self.assertEqual([], gc.subdirectories())
