@@ -31,14 +31,27 @@ import jinja2
 import logging
 import os
 import posixpath
+from typing import (
+    List,
+    Tuple,
+    Iterable,
+    Iterator,
+    Optional,
+)
+
 import shutil
 import urllib.parse
 
 from xandikos import __version__ as xandikos_version
 from xandikos import (access, apache, caldav, carddav, quota, sync, webdav,
                       infit, scheduling, timezones)
-from xandikos.icalendar import ICalendarFile
+from xandikos.icalendar import (
+    ICalendarFile,
+    CalendarFilter,
+)
 from xandikos.store import (
+    Store,
+    File,
     DuplicateUidError,
     InvalidFileContents,
     NoSuchItem,
@@ -48,6 +61,7 @@ from xandikos.store import (
     STORE_TYPE_PRINCIPAL,
     STORE_TYPE_SCHEDULE_INBOX,
     STORE_TYPE_SCHEDULE_OUTBOX,
+    STORE_TYPE_SUBSCRIPTION,
     STORE_TYPE_OTHER,
 )
 from xandikos.store.git import (
@@ -66,27 +80,34 @@ CALENDAR_HOME_SET = ['calendars']
 ADDRESSBOOK_HOME_SET = ['contacts']
 
 TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'templates')
-jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATES_DIR))
+jinja_env = jinja2.Environment(
+    loader=jinja2.FileSystemLoader(TEMPLATES_DIR),
+    enable_async=True)
 
 
-def render_jinja_page(name, accepted_content_languages, **kwargs):
+async def render_jinja_page(
+        name: str, accepted_content_languages: List[str],
+        **kwargs) -> Tuple[
+            Iterable[bytes], int, Optional[str], str, List[str]]:
     """Render a HTML page from jinja template.
 
     :param name: Name of the page
     :param accepted_content_languages: List of accepted content languages
-    :return: TUple of (body, content_length, etag, content_type, languages)
+    :return: Tuple of (body, content_length, etag, content_type, languages)
     """
     # TODO(jelmer): Support rendering other languages
     encoding = 'utf-8'
     template = jinja_env.get_template(name)
-    body = template.render(
+    body = await template.render_async(
         version=xandikos_version,
-        urljoin=urllib.parse.urljoin, **kwargs).encode(encoding)
-    return ([body], len(body), None, 'text/html; encoding=%s' % encoding,
+        urljoin=urllib.parse.urljoin, **kwargs)
+    body_encoded = body.encode(encoding)
+    return ([body_encoded], len(body_encoded),
+            None, 'text/html; encoding=%s' % encoding,
             ['en-UK'])
 
 
-def create_strong_etag(etag):
+def create_strong_etag(etag: str) -> str:
     """Create strong etags.
 
     :param etag: basic etag
@@ -95,7 +116,7 @@ def create_strong_etag(etag):
     return '"' + etag + '"'
 
 
-def extract_strong_etag(etag):
+def extract_strong_etag(etag: Optional[str]) -> Optional[str]:
     """Extract a strong etag from a string."""
     if etag is None:
         return etag
@@ -105,27 +126,28 @@ def extract_strong_etag(etag):
 class ObjectResource(webdav.Resource):
     """Object resource."""
 
-    def __init__(self, store, name, content_type, etag):
+    def __init__(self, store: Store, name: str, content_type: str, etag: str,
+                 file: Optional[File] = None):
         self.store = store
         self.name = name
         self.etag = etag
         self.content_type = content_type
-        self._file = None
+        self._file = file
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "%s(%r, %r, %r, %r)" % (
             type(self).__name__, self.store, self.name, self.etag,
             self.get_content_type()
         )
 
     @property
-    def file(self):
+    def file(self) -> File:
         if self._file is None:
             self._file = self.store.get_file(self.name, self.content_type,
                                              self.etag)
         return self._file
 
-    def get_body(self):
+    async def get_body(self) -> Iterable[bytes]:
         return self.file.content
 
     def set_body(self, data, replace_etag=None):
@@ -144,16 +166,16 @@ class ObjectResource(webdav.Resource):
                 'UID already in use.')
         return create_strong_etag(etag)
 
-    def get_content_language(self):
+    def get_content_language(self) -> str:
         raise KeyError
 
-    def get_content_type(self):
+    def get_content_type(self) -> str:
         return self.content_type
 
-    def get_content_length(self):
-        return sum(map(len, self.get_body()))
+    async def get_content_length(self) -> int:
+        return sum(map(len, await self.get_body()))
 
-    def get_etag(self):
+    async def get_etag(self) -> str:
         return create_strong_etag(self.etag)
 
     def get_supported_locks(self):
@@ -222,51 +244,58 @@ class StoreBasedCollection(object):
             self.store.set_type(STORE_TYPE_SCHEDULE_OUTBOX)
         elif resource_types == {webdav.COLLECTION_RESOURCE_TYPE}:
             self.store.set_type(STORE_TYPE_OTHER)
+        elif resource_types == {webdav.COLLECTION_RESOURCE_TYPE,
+                                caldav.SUBSCRIPTION_RESOURCE_TYPE}:
+            self.store.set_type(STORE_TYPE_SUBSCRIPTION)
         else:
             raise NotImplementedError(self.set_resource_types)
 
-    def _get_resource(self, name, content_type, etag):
-        return ObjectResource(self.store, name, content_type, etag)
+    def _get_resource(self, name: str,
+                      content_type: str,
+                      etag: str,
+                      file: Optional[File] = None) -> webdav.Resource:
+        return ObjectResource(self.store, name, content_type, etag, file=file)
 
-    def _get_subcollection(self, name):
+    def _get_subcollection(self, name: str) -> webdav.Collection:
         return self.backend.get_resource(posixpath.join(self.relpath, name))
 
-    def get_displayname(self):
+    def get_displayname(self) -> str:
         displayname = self.store.get_displayname()
         if displayname is None:
             return os.path.basename(self.store.repo.path)
         return displayname
 
-    def set_displayname(self, displayname):
+    def set_displayname(self, displayname: str) -> None:
         self.store.set_displayname(displayname)
 
-    def get_sync_token(self):
+    def get_sync_token(self) -> str:
         return self.store.get_ctag()
 
-    def get_ctag(self):
+    def get_ctag(self) -> str:
         return self.store.get_ctag()
 
-    def get_etag(self):
+    async def get_etag(self) -> str:
         return create_strong_etag(self.store.get_ctag())
 
-    def members(self):
-        ret = []
+    def members(self) -> Iterator[Tuple[str, webdav.Resource]]:
         for (name, content_type, etag) in self.store.iter_with_etag():
             resource = self._get_resource(name, content_type, etag)
-            ret.append((name, resource))
+            yield (name, resource)
+        for (name, resource) in self.subcollections():
+            yield (name, resource)
+
+    def subcollections(self):
         for name in self.store.subdirectories():
-            ret.append((name, self._get_subcollection(name)))
-        return ret
+            yield (name, self._get_subcollection(name))
 
     def get_member(self, name):
         assert name != ''
         for (fname, content_type, fetag) in self.store.iter_with_etag():
             if name == fname:
                 return self._get_resource(name, content_type, fetag)
-        else:
-            if name in self.store.subdirectories():
-                return self._get_subcollection(name)
-            raise KeyError(name)
+        if name in self.store.subdirectories():
+            return self._get_subcollection(name)
+        raise KeyError(name)
 
     def delete_member(self, name, etag=None):
         assert name != ''
@@ -277,7 +306,8 @@ class StoreBasedCollection(object):
             # self.get_subcollection(name).destroy()
             shutil.rmtree(os.path.join(self.store.path, name))
 
-    def create_member(self, name, contents, content_type):
+    def create_member(self, name: str, contents: Iterable[bytes],
+                      content_type: str) -> Tuple[str, str]:
         try:
             (name, etag) = self.store.import_one(name, content_type, contents)
         except InvalidFileContents as e:
@@ -291,7 +321,11 @@ class StoreBasedCollection(object):
                 'UID already in use.')
         return (name, create_strong_etag(etag))
 
-    def iter_differences_since(self, old_token, new_token):
+    def iter_differences_since(
+            self, old_token: str, new_token: str) -> Iterator[Tuple[
+                str, Optional[webdav.Resource], Optional[webdav.Resource]]]:
+        old_resource: Optional[webdav.Resource]
+        new_resource: Optional[webdav.Resource]
         for (name, content_type,
              old_etag, new_etag) in self.store.iter_changes(
                  old_token, new_token):
@@ -337,26 +371,26 @@ class StoreBasedCollection(object):
     def get_content_language(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
-    def destroy(self):
+    def destroy(self) -> None:
         # RFC2518, section 8.6.2 says this should recursively delete.
         self.store.destroy()
 
-    def get_body(self):
+    async def get_body(self):
         raise NotImplementedError(self.get_body)
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'collection.html', accepted_content_languages, collection=self,
             self_url=self_url)
 
-    def get_is_executable(self):
+    def get_is_executable(self) -> bool:
         return False
 
     def get_quota_used_bytes(self):
@@ -388,6 +422,35 @@ class ScheduleOutbox(StoreBasedCollection, scheduling.ScheduleOutbox):
     """A schedling outbox collection."""
 
 
+class SubscriptionCollection(StoreBasedCollection, caldav.Subscription):
+
+    def get_source_url(self):
+        source_url = self.store.get_source_url()
+        if source_url is None:
+            raise KeyError
+        return source_url
+
+    def set_source_url(self, url):
+        self.store.set_source_url(url)
+
+    def get_calendar_description(self):
+        return self.store.get_description()
+
+    def get_calendar_color(self):
+        color = self.store.get_color()
+        if not color:
+            raise KeyError
+        if color and color[0] != '#':
+            color = '#' + color
+        return color
+
+    def set_calendar_color(self, color):
+        self.store.set_color(color)
+
+    def get_supported_calendar_components(self):
+        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY"]
+
+
 class CalendarCollection(StoreBasedCollection, caldav.Calendar):
 
     def get_calendar_description(self):
@@ -404,8 +467,17 @@ class CalendarCollection(StoreBasedCollection, caldav.Calendar):
     def set_calendar_color(self, color):
         self.store.set_color(color)
 
+    def get_calendar_order(self):
+        order = self.store.config.get_order()
+        if not order:
+            raise KeyError
+        return order
+
+    def set_calendar_order(self, order):
+        self.store.config.set_order(order)
+
     def get_calendar_timezone(self):
-        # TODO(jelmer): Read a magic file from the store?
+        # TODO(jelmer): Read from config
         raise KeyError
 
     def set_calendar_timezone(self, content):
@@ -450,6 +522,14 @@ class CalendarCollection(StoreBasedCollection, caldav.Calendar):
         # TODO(jelmer)
         raise KeyError
 
+    def calendar_query(self, create_filter_fn):
+        filter = create_filter_fn(CalendarFilter)
+        for (name, file, etag) in self.store.iter_with_filter(
+                filter=filter):
+            resource = self._get_resource(
+                name, file.content_type, etag, file=file)
+            yield (name, resource)
+
 
 class AddressbookCollection(StoreBasedCollection, carddav.Addressbook):
 
@@ -503,7 +583,7 @@ class CollectionSetResource(webdav.Collection):
     def get_sync_token(self):
         raise KeyError
 
-    def get_etag(self):
+    async def get_etag(self):
         raise KeyError
 
     def get_ctag(self):
@@ -516,14 +596,12 @@ class CollectionSetResource(webdav.Collection):
         return []
 
     def members(self):
-        ret = []
         p = self.backend._map_to_file_path(self.relpath)
         for name in os.listdir(p):
             if name.startswith('.'):
                 continue
             resource = self.get_member(name)
-            ret.append((name, resource))
-        return ret
+            yield (name, resource)
 
     def get_member(self, name):
         assert name != ''
@@ -548,7 +626,7 @@ class CollectionSetResource(webdav.Collection):
     def get_content_language(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
     def get_last_modified(self):
@@ -564,12 +642,12 @@ class CollectionSetResource(webdav.Collection):
         # RFC2518, section 8.6.2 says this should recursively delete.
         shutil.rmtree(p)
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'root.html', accepted_content_languages, self_url=self_url)
 
     def get_is_executable(self):
@@ -587,7 +665,7 @@ class CollectionSetResource(webdav.Collection):
 class RootPage(webdav.Resource):
     """A non-DAV resource."""
 
-    resource_types = []
+    resource_types: List[str] = []
 
     def __init__(self, backend):
         self.backend = backend
@@ -602,10 +680,10 @@ class RootPage(webdav.Resource):
             principals=self.backend.find_principals(),
             self_url=self_url)
 
-    def get_body(self):
+    async def get_body(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
     def get_content_type(self):
@@ -617,9 +695,9 @@ class RootPage(webdav.Resource):
     def get_active_locks(self):
         return []
 
-    def get_etag(self):
+    async def get_etag(self):
         h = hashlib.md5()
-        for c in self.get_body():
+        for c in await self.get_body():
             h.update(c)
         return h.hexdigest()
 
@@ -630,11 +708,11 @@ class RootPage(webdav.Resource):
         return ['en-UK']
 
     def get_member(self, name):
-        return self.backend.get_resource(name)
+        return self.backend.get_resource('/' + name)
 
     def delete_member(self, name, etag=None):
         # This doesn't have any non-collection members.
-        self.get_member(name).destroy()
+        self.get_member('/' + name).destroy()
 
     def get_is_executable(self):
         return False
@@ -653,6 +731,9 @@ class Principal(webdav.Principal):
     def get_principal_url(self):
         return '.'
 
+    def get_principal_address(self):
+        raise KeyError
+
     def get_calendar_home_set(self):
         return CALENDAR_HOME_SET
 
@@ -710,6 +791,9 @@ class Principal(webdav.Principal):
         # TODO(jelmer): make this configurable
         return 'inbox'
 
+    def get_creationdate(self):
+        raise KeyError
+
 
 class PrincipalBare(CollectionSetResource, Principal):
     """Principal user resource."""
@@ -729,15 +813,19 @@ class PrincipalBare(CollectionSetResource, Principal):
                 pass
         return p
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'principal.html', accepted_content_languages, principal=self,
             self_url=self_url)
 
+    def subcollections(self):
+        # TODO(jelmer): Return members
+        return []
+
 
 class PrincipalCollection(Collection, Principal):
     """Principal user resource."""
@@ -796,6 +884,8 @@ class XandikosBackend(webdav.Backend):
 
     def get_resource(self, relpath):
         relpath = posixpath.normpath(relpath)
+        if not relpath.startswith('/'):
+            raise ValueError('relpath %r should start with /')
         if relpath == '/':
             return RootPage(self)
         p = self._map_to_file_path(relpath)
@@ -815,6 +905,7 @@ class XandikosBackend(webdav.Backend):
                     STORE_TYPE_PRINCIPAL: PrincipalCollection,
                     STORE_TYPE_SCHEDULE_INBOX: ScheduleInbox,
                     STORE_TYPE_SCHEDULE_OUTBOX: ScheduleOutbox,
+                    STORE_TYPE_SUBSCRIPTION: SubscriptionCollection,
                     STORE_TYPE_OTHER: Collection
                 }[store.get_type()](self, relpath, store)
         else:
@@ -853,10 +944,12 @@ class XandikosApp(webdav.WebDAVApp):
             webdav.GetContentTypeProperty(),
             webdav.GetContentLengthProperty(),
             webdav.GetContentLanguageProperty(),
+            caldav.SourceProperty(),
             caldav.CalendarHomeSetProperty(),
             carddav.AddressbookHomeSetProperty(),
             caldav.CalendarDescriptionProperty(),
             caldav.CalendarColorProperty(),
+            caldav.CalendarOrderProperty(),
             caldav.SupportedCalendarComponentSetProperty(),
             carddav.AddressbookDescriptionProperty(),
             carddav.PrincipalAddressProperty(),
@@ -917,24 +1010,6 @@ class XandikosApp(webdav.WebDAVApp):
         ])
 
 
-class WellknownRedirector(object):
-    """Redirect paths under .well-known/ to the appropriate paths."""
-
-    def __init__(self, inner_app, dav_root):
-        self._inner_app = inner_app
-        self._dav_root = dav_root
-
-    def __call__(self, environ, start_response):
-        # See https://tools.ietf.org/html/rfc6764
-        path = posixpath.normpath(
-            environ['SCRIPT_NAME'] + environ['PATH_INFO'])
-        if path in WELLKNOWN_DAV_PATHS:
-            start_response('302 Found', [
-                ('Location', self._dav_root)])
-            return []
-        return self._inner_app(environ, start_response)
-
-
 def create_principal_defaults(backend, principal):
     """Create default calendar and addressbook for a principal.
 
@@ -972,6 +1047,16 @@ def create_principal_defaults(backend, principal):
         logging.info('Create inbox in %s.', resource.store.path)
 
 
+class RedirectDavHandler(object):
+
+    def __init__(self, dav_root: str):
+        self._dav_root = dav_root
+
+    async def __call__(self, request):
+        from aiohttp import web
+        return web.HTTPFound(self._dav_root)
+
+
 def main(argv):
     import argparse
     import sys
@@ -987,8 +1072,9 @@ def main(argv):
 
     access_group = parser.add_argument_group(title="Access Options")
     access_group.add_argument(
-        "-l", "--listen_address", dest="listen_address", default="localhost",
-        help="Binding IP address. [%(default)s]")
+        "-l", "--listen-address", dest="listen_address", default="localhost",
+        help=("Bind to this address. "
+              "Pass in path for unix domain socket. [%(default)s]"))
     access_group.add_argument(
         "-p", "--port", dest="port", type=int, default=8080,
         help="Port to listen on. [%(default)s]")
@@ -1046,28 +1132,50 @@ def main(argv):
             'Run xandikos with --autocreate?',
             options.current_user_principal)
 
-    app = XandikosApp(
+    main_app = XandikosApp(
         backend,
         current_user_principal=options.current_user_principal)
 
-    from wsgiref.simple_server import make_server
-    app = WellknownRedirector(app, options.route_prefix)
-    server = make_server(options.listen_address, options.port, app)
-    logging.info('Listening on %s:%s', options.listen_address,
-                 options.port)
+    async def xandikos_handler(request):
+        return await main_app.aiohttp_handler(request, options.route_prefix)
 
-    import signal
+    if '/' in options.listen_address:
+        socket_path = options.listen_address
+        listen_address = None
+        logging.info('Listening on unix domain socket %s', socket_path)
+    else:
+        listen_address = options.listen_address
+        socket_path = None
+        logging.info('Listening on %s:%s', listen_address, options.port)
 
-    def handle_sigterm(sig, action):
-        sys.exit(0)
+    from aiohttp import web
 
-    signal.signal(signal.SIGTERM, handle_sigterm)
+    app = web.Application()
     try:
-        server.serve_forever()
-    except KeyboardInterrupt:
-        pass
-    finally:
-        server.shutdown()
+        import prometheus_client  # noqa: F401
+    except ModuleNotFoundError:
+        logging.warning(
+            'Prometheus client not found; /metrics will not be available.')
+    else:
+        from .metrics import setup_metrics
+        setup_metrics(app)
+
+    for path in WELLKNOWN_DAV_PATHS:
+        app.router.add_route(
+            "*", path,
+            RedirectDavHandler(options.route_prefix).__call__)
+
+    if options.route_prefix.strip('/'):
+        xandikos_app = web.Application()
+        xandikos_app.router.add_route("*", "/{path_info:.*}", xandikos_handler)
+
+        async def redirect_to_subprefix(request):
+            return web.HTTPFound(options.route_prefix)
+        app.router.add_route("*", "/", redirect_to_subprefix)
+        app.add_subapp(options.route_prefix, xandikos_app)
+    else:
+        app.router.add_route("*", "/{path_info:.*}", xandikos_handler)
+    web.run_app(app, port=options.port, host=listen_address, path=socket_path)
 
 
 if __name__ == '__main__':
